#!/usr/bin/env python

from __future__ import print_function

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"


"""
    HTTP connection pool with LIFO logic, and 2 implementations of connections: pycurl (fast) and httplib (slow).
    HTTPS is supported.
"""
import threading
import socket
import re
import logging
import six
import pycurl

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

if six.PY2:
    import httplib
    from StringIO import StringIO as BytesIO
else:
    import http.client as httplib
    from io import BytesIO

try:
    # Python < 2.6 doesn't verify SSL certification
    # disable SSL certificate validation on python >= 2.7.9 as well
    import ssl
    ssl._create_default_https_context = ssl._create_unverified_context
except ImportError:
    pass


class HTTPConnectionPycurl:
    def __init__(self, prefix, key_file=None, cert_file=None):
        self.prefix = prefix
        self.curl = pycurl.Curl()
        self.curl.setopt(pycurl.SSL_VERIFYPEER, 0)
        self.curl.setopt(pycurl.SSL_VERIFYHOST, 0)
        if key_file:
            self.curl.setopt(pycurl.SSLKEY, key_file)
        if cert_file:
            self.curl.setopt(pycurl.SSLCERT, cert_file)
        self.curl.setopt(pycurl.ENCODING, "")
        self.buf = None
        self.status = 0
        self.response_headers = None
        self._response_string = None
        self.reason = ''
        self.cleaning_needed = False

    def close(self):
        self.curl.close()

    def request(self, verb, path, body, headers):
        c = self.curl
        hdrs = [str(h + ": " + v) for h, v in six.iteritems(headers)] if headers else []
        verb = verb.upper()
        if verb == 'GET':
            if self.cleaning_needed:
                c.setopt(pycurl.POST, 0)
                c.unsetopt(pycurl.CUSTOMREQUEST)
                c.setopt(pycurl.NOBODY, 0)
                self.cleaning_needed = False
            if body:
                self.cleaning_needed = True
                c.setopt(pycurl.POST, 0)
                c.setopt(pycurl.CUSTOMREQUEST, verb)
                c.setopt(pycurl.NOBODY, 0)
                c.setopt(pycurl.POSTFIELDS, body or "")
        elif verb == 'POST':
            self.cleaning_needed = True
            c.unsetopt(pycurl.CUSTOMREQUEST)
            c.setopt(pycurl.NOBODY, 0)
            c.setopt(pycurl.POST, 1)
            c.setopt(pycurl.POSTFIELDS, body or "")
            hdrs.append("Expect:")
        elif verb == 'PUT' or verb == "DELETE":
            self.cleaning_needed = True
            c.setopt(pycurl.POST, 0)
            c.setopt(pycurl.CUSTOMREQUEST, verb)
            c.setopt(pycurl.NOBODY, 0)
            c.setopt(pycurl.POSTFIELDS, body or "")
        elif verb == 'HEAD':
            self.cleaning_needed = True
            c.setopt(pycurl.POST, 0)
            c.unsetopt(pycurl.CUSTOMREQUEST)
            c.setopt(pycurl.NOBODY, 1)
        else:
            raise pycurl.error("unsupported verb: " + verb)
        c.setopt(pycurl.URL, str(self.prefix + path))
        c.setopt(pycurl.HTTPHEADER, hdrs)
        self.buf = BytesIO()
        self.response_headers = []
        c.setopt(pycurl.WRITEFUNCTION, self.buf.write)
        c.setopt(pycurl.HEADERFUNCTION, self._header_handler)
        c.perform()

    def getresponse(self):
        self.status = self.curl.getinfo(pycurl.HTTP_CODE)
        m = re.match(r'HTTP\/\S*\s*\d+\s*(.*?)\s*$', self._response_string)
        if m:
            self.reason = m.group(1)
        else:
            self.reason = ''
        return self

    def set_debuglevel(self, level):
        self.curl.setopt(pycurl.VERBOSE, 1 if level else 0)
        if level:
            self.curl.setopt(pycurl.DEBUGFUNCTION, self._debug)

    def _debug(self, debug_type, debug_msg):
        if debug_type in (0, 3, 4, 5, 6):
            return  # skip details(0), body (3,4), and ssl (5,6)
        if type(debug_msg) == bytes:
            debug_msg = debug_msg.decode("utf-8")
        if debug_type == 2:
            debug_msg = debug_msg.replace("\r\n", "\n")
        logging.log(logging.DEBUG - 1, "pycurl(%d): %s" % (debug_type, debug_msg.strip()))

    def read(self):
        return self.buf.getvalue()

    def getheaders(self):
        return self.response_headers

    def getheader(self, header, default=None):
        for h, v in self.response_headers:
            if h.lower() == header.lower():
                return v
        return default

    def _header_handler(self, line):
        if type(line) == bytes:
            # HTTP standard specifies that headers are encoded in iso-8859-1.
            line = line.decode('iso-8859-1')
        x = line.split(':', 1)
        if len(x) != 2:
            # check status message
            if line.startswith("HTTP/"):
                self._response_string = line
            return
        self.response_headers.append((x[0].strip(), x[1].strip()))


class _ctx_manager:
    def __init__(self, pool):
        self.pool = pool

    def __enter__(self):
        self.tup = self.pool.get()
        return self.tup

    def __exit__(self, exc_type, exc_value, traceback):
        con = self.tup[0]
        if exc_type is not None or getattr(con, "_invalidate", False):
            self.pool.cnt -= 1
            self.pool._dctor(con)
            # logging.error("!!!!pool.borrow.exit - error. cnt=%d, inpool=%d", pool.cnt, len(pool._items))
        else:
            self.pool.put(con)
            # logging.error("!!!!pool.borrow.exit - ok. cnt=%d, inpool=%d", pool.cnt, len(pool._items))


class LIFOPool(object):
    def __init__(self, ctor, dctor, max_items=10, verbose=None):
        self._ctor = ctor
        self._dctor = dctor
        self.max_items = max(0, max_items)
        self.verbose = logging.getLogger().getEffectiveLevel() < logging.DEBUG if verbose is None else verbose
        self._items = []
        self._lock = threading.Lock()
        self.cnt = 0

    def __del__(self):
        self.clear()

    def clear(self):
        """Clear all pool"""
        with self._lock:
            while self._items:
                self.cnt -= 1
                self._dctor(self._items.pop())

    def get(self):
        """Get an (item, is_new) from the pool or create a new one. After use, return item via put().
        Or, better, use borrow() that ensures the item is properly returned"""
        with self._lock:
            if self._items:
                return self._items.pop(), False
            self.cnt += 1
        con = self._ctor()
        if self.verbose:
            con.set_debuglevel(1)
        return con, True

    def put(self, item):
        """Return item, previously obtained via get(). Do not return bad (not reusable) items.
        Or, better, use borrow() that ensures the item is properly returned"""
        with self._lock:
            self._items.append(item)
            while len(self._items) > self.max_items:
                self.cnt -= 1
                self._dctor(self._items.pop(0))

    def borrow(self):
        """Exception-safe fool-proof way to get and return the (item, is_new). Use with keyword 'with' like this:
        with pool.borrow() as (x, is_new):
            do_stuff(x)
        """
        return _ctx_manager(self)


class HTTPPool(LIFOPool):
    """Connection pool, keeps at most <max_conns> connections to given <server_uri>.
    Parses URI and uses appropriate HTTP/HTTPS connection objects.
    If <max_items> == 0, it is equivalent to keep-alive = False
    Connections are handled in LIFO order, thread safety is provided.
    Public properties (readonly):
        host - server host (from URI)
        port - server port (from URI)
        url - full server part of uri, like "https://yourserver.com:1234"
    """
    def __init__(self, server_uri, max_conns=10, parse_exception=Exception, key_file=None, cert_file=None,
                 verbose=None, engine="pycurl"):
        u = urlparse(server_uri)
        if u.params != '':
            raise parse_exception("Invalid URI: " + server_uri)

        if u.netloc.find(":") < 0:
            self.host = u.netloc
            self.port = {'http': 80, 'https': 443}.get(u.scheme)
        else:
            self.host, self.port = u.netloc.split(":")
            self.port = int(self.port)

        self.url = server_uri

        if engine == "pycurl":
            def ctor():
                return HTTPConnectionPycurl(server_uri, key_file=key_file, cert_file=cert_file)
            self.temporary_errors = (pycurl.error, socket.error)
            self.fatal_errors = (pycurl.error, )   # we don't know suitable errors
        elif engine == "httplib":
            if u.scheme == 'https':
                def ctor():
                    return httplib.HTTPSConnection(self.host, self.port, key_file=key_file, cert_file=cert_file)
            else:
                def ctor():
                    return httplib.HTTPConnection(self.host, self.port)
            self.temporary_errors = (httplib.BadStatusLine, httplib.ImproperConnectionState, socket.error)
            self.fatal_errors = (httplib.HTTPException, )
        else:
            raise Exception("unknown http engine")

        LIFOPool.__init__(self, ctor=ctor, dctor=lambda x: x.close(), max_items=max_conns, verbose=verbose)


##############################################################################
# Autotests
##############################################################################


def _coverage():
    p = HTTPPool("127.0.0.1")
    print("OK")


if __name__ == "__main__":
    _coverage()



#!/usr/bin/env python

from __future__ import print_function

import os
import sys
import time
from tempfile import gettempdir

bindir, basename = os.path.split(os.path.abspath(__file__))
sys.path.append(os.path.join(bindir, ".."))

basename = basename.split(".")[0]

from optparse import OptionParser, OptionGroup, IndentedHelpFormatter
from perftracker_cp_crawler.cp_crawler import CPCrawler, CPCrawlerException
from perftracker_cp_crawler.wpa_cp_engine import CPWPAdminConsole
from distutils.version import LooseVersion

# from perftracker_cp_crawler import __version__ as perftrack_cp_crawler_lib_version
# PERFTRACKER_CP_CRAWLER_VERSION_REQUIRED = '0.0.5'
# if LooseVersion(perftrackerlib_version) < LooseVersion(PERFTRACKER_CP_CRAWLER_VERSION_REQUIRED):
#    print("Error: perftracker_cp_crawler version >= %s must be installed, found %s" %
#          (PERFTRACKER_CP_CRALER_VERSION_REQUIRED, perftrackerlib_version))
#    sys.exit(-2)
#
# see perftracker_cp_crawler/wpa_cp_engine.py as reference implementation
#

def main():
    usage = "usage: %prog [options] URL [URL2 [URL3 ...]]"

    description = "Example: -m -U user -P user https://demo.wpjobboard.net/wp-login.php"

    workdir = os.path.join(gettempdir(), "%s.%d" % (basename, os.getpid()))
    logfile = os.path.join(workdir, basename + ".log")

    cpc = CPCrawler(workdir=workdir, logfile=logfile)

    op = OptionParser(usage=usage, description=description, formatter=IndentedHelpFormatter(width=120))
    cpc.add_options(op, passwd='pass', ajax_threshold=0.5)

    opts, args = op.parse_args()

    if not args:
        op.error("URL is not specified, %s" % description)
        sys.exit(-1)

    cpc.init_opts(opts, args)

    cpc.crawl([CPWPAdminConsole])


if __name__ == "__main__":
    main()



#!/usr/bin/env python

from __future__ import print_function, absolute_import

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

"""
A library to generate HTML report with pages information
"""

import os
from collections import OrderedDict
from PIL import Image

from perftrackerlib import __version__ as __version__
from perftrackerlib.helpers.timeline import ptDoc, ptSection
from .page import Page


class ptBrowserHtmlReport:
    def __init__(self, output_dir, title=''):
        self.pages = OrderedDict()
        self._url2id = {}
        self.output_dir = output_dir
        self.title = title
        self.doc = ptDoc(title=title)

        self.sdir = os.path.join(self.output_dir, 'screenshots')
        if not os.path.exists(self.sdir):
            os.makedirs(self.sdir)

    def add_page(self, url, title=None, page=None, render_html=True):
        if self.pages.get(url, None) is None:
            self.pages[url] = []
        if not page and title:
            page = Page(None, url, name=title)
        if page:
            self.pages[url].append(page)
        if render_html:
            self.gen_index_html()
        return self.get_screenshot_file_path(url)

    def url2id(self, url):
        if url not in self._url2id:
            self._url2id[url] = len(self._url2id)
        return self._url2id[url]

    def get_screenshot_file_path(self, url, sfx=""):
        return os.path.join(self.sdir, '%d%s.png' % (self.url2id(url), str(sfx)))

    def gen_thumbnails(self, url):
        if url not in self._url2id:
            return

        p = self.get_screenshot_file_path(url)
        i = Image.open(p)
        for w in 1024, 256:
            ratio = w / float(i.size[0])
            h = int((float(i.size[1]) * ratio))
            img = i.resize((w, h), Image.ANTIALIAS)
            img.save(self.get_screenshot_file_path(url, "_%d" % w))
            img.close()

        i.close()
        # os.unlink(p)

    def render_html(self):

        html = """
<!DOCTYPE html>
<html lang='en'>
<head>
  <title>%s</title>
  <meta charset='utf-8'>
  <meta name='viewport' content='width=device-width, initial-scale=1'>
  <link rel='stylesheet' href='https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css'>
  <script src='https://ajax.googleapis.com/ajax/libs/jquery/3.3.1/jquery.min.js'></script>
  <script src='https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/js/bootstrap.min.js'></script>
<style type='text/css'>
.modal-lg { max-width: 1050px !important; width: 90%% !important; }
.modal-lg img { width: 100%%; }
</style>
</head>
<body>
<div class='jumbotron text-center'><h2>Pages</h2></div>
<div class='container'>
""" % (str(self.title))

        for url, pages in self.pages.items():
            img_256 = self.get_screenshot_file_path(url, sfx="_256")
            img_1024 = self.get_screenshot_file_path(url, sfx="_1024")

            if len(pages) and pages[-1] and pages[-1].name:
                page_name = pages[-1].name
            else:
                page_name = ""

            if len(pages) and pages[-1] and pages[-1].ts_start:
                p = pages[-1]
                reqs = p.get_uncached_reqs()
                details = []
                details.append("%d requests" % len(reqs))
                details.append("%d msec" % int(p.dur))
                details.append("%.0f KB" % (sum([r.length for r in reqs]) / 1024.0))
                details.append("page is not tested yet")
            else:
                details = ["page is not tested yet"]

            html += "<div class='row'>"

            html += "<div class='col-sm-8'>"
            html += "<h3>%s</h3><p><a href='%s'>%s</a><br>%s</p>" % (page_name, url, url, "<br>".join(details))
            html += "</div>"

            html += "<div class='col-sm-4'>"
            html += "<a data-toggle='modal' data-target='#modal%d'><img src='%s'></a>" % (self.url2id(url), img_256)
            html += "</div>"

            html += "</div>"

            html += """
  <!-- Modal -->
  <div class="modal" id="modal%d" role="dialog">
    <div class="modal-dialog modal-lg">
      <div class="modal-content">
        <div class="modal-header">
          <button type="button" class="close" data-dismiss="modal">&times;</button>
          <h4 class="modal-title">%s</h4>
        </div>
        <div class="modal-body">
          <img src='%s'>
        </div>
      </div>
    </div>
  </div>
""" % (self.url2id(url), page_name, img_1024)

        html += "</div>"
        html += "</body></html>"
        return html

    def gen_index_html(self):
        p = os.path.join(self.output_dir, 'index.html')
        f = open(p, 'w')
        f.write(self.render_html())
        f.close()
        return p


##############################################################################
# Autotests
##############################################################################


if __name__ == "__main__":
    r = ptBrowserHtmlReport('')
    r.add_page('', None)
    r.render_html()
    print("OK")



#!/usr/bin/env python

from __future__ import print_function

# -*- coding: utf-8 -*-
__author__ = "istillc0de@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

import re
import logging
import urllib
import tempfile
import os
import sys

if sys.version_info[0] < 3:
    import urlparse
    import httplib
    from httplib import HTTPResponse
    from BaseHTTPServer import BaseHTTPRequestHandler
    from StringIO import StringIO
    from HTMLParser import HTMLParser
else:
    import urllib.parse as urlparse
    import http.client as httplib
    from http.client import HTTPResponse
    from http.server import BaseHTTPRequestHandler
    from io import StringIO
    from html.parser import HTMLParser


reScriptLocation = re.compile(r"top\.location[\.href]*=\'(.*)\'")
reFormButtonUrl = re.compile(r"doSubmit\(\'(.*)\'")

###############################################################################
# Support classes for HTML page processing and action URLs extraction
#
class TagValueBase(object):
    def __init__(self, tag):
        self.tag = tag
        self.value = None
        self.inside = False

    def is_processing(self):
        return self.inside

class TagValueManual(TagValueBase):
    def __init__(self, tag):
        TagValueBase.__init__(self, tag)

    def begin(self, tag):
        if tag == self.tag:
            self.inside = True
            return True
        return False

    def end(self, tag):
        if tag == self.tag:
            self.inside = False
            return True
        return False

    def set_value(self, value):
        self.value = value

    def clear_value(self):
        self.value = None

class Request(object):
    def __init__(self, url, method=None, uid=None):
        self._params = {}              # dict of parameters {name: value} (used by POST only)

        self.url = None               # URL without host name, e.g. /turbine
        self.method = method.upper() if method else 'GET'
        self.uid = uid                # request ID

        if url:  # if url is None, Request considered as valid but incomplete
            self.reset_url(url)

    def reset_url(self, url):
        url = urlparse.urlparse(url)
        self.url = url.path

        if not self.url.startswith('/'):
            self.url = '/' + self.url

        if self.method == 'POST':
            self._params = dict(urlparse.parse_qsl(url.query))
        else:
            self._params = {}
            if url.query:
                self.url += '?' + url.query

    def add_param(self, name, value):
        if self.method == 'POST':
            self._params[name] = value
        else:
            self.url += '&%s=%s' % (name, value)

    def get_params(self):
        # converts parameters dict into string (n1=v1&n2=v2&...)
        return urllib.urlencode(self._params) if self._params else None

    def is_complete(self):
        return True if self.url else False

    def process_action(self, tag, attrs):
        if tag == 'button' and attrs.get('type','submit') == 'submit':
            match = reFormButtonUrl.search(attrs.get('onclick', ''))
            if match:
                url = urllib.unquote(match.group(1))
                url = HTMLParser().unescape(url)
                self.uid = tag + ':' + attrs.get('name', 'none')
                self.reset_url(url)
                return True

        return False


###############################################################################
# Class for HTML page processing (based on HTMLParser template)
#
# The main goal of class is gather all possible URLs of actions on page. After
# that, user would select one and use it either to build the chain of auto-redirected
# pages or for manual navigation (JFI: list of possible redirection methods
# https://code.google.com/p/html5security/wiki/RedirectionMethods).
#
class PageParser(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)

        self._body_tag = TagValueManual('body')
        self._form_tag = TagValueManual('form')
        self._script_tag = TagValueManual('script')
        self._noscript_tag = TagValueManual('noscript')

        # Objects of Request type (public)
        self.forms = []
        self.iframe_src = None
        self.body_onload = None
        self.script = None
        self.global_action = None  # action outside <form>'s

        self.form_auto_submit = False

    def handle_starttag(self, tag, attributes):
        attrs = {}
        for attr in attributes:
            attrs[attr[0].lower()] = attr[1]

        # <noscript>  (there is no any valuable information inside this tag)
        if self._noscript_tag.begin(tag) or self._noscript_tag.is_processing():
            return

        # <script>  (only set the flag, script body will be in handle_data())
        if self._script_tag.begin(tag):
            return

        # <iframe>  (process 'src' with URL)
        if tag == 'iframe':
            attr = attrs.get('src')
            if attr:
                self.iframe_src = Request(attr)
            return

        # <body>  (process 'onload' with script)
        if self._body_tag.begin(tag):
            attr = attrs.get('onload')
            if not attr:
                pass
            elif "document.forms[0].submit()" in attr:
                self.form_auto_submit = True
            else:
                match = reScriptLocation.search(attr)
                if match:
                    url = urllib.unquote(match.group(1))
                    url = HTMLParser().unescape(url)
                    self.body_onload = Request(url)
            self._body_tag.set_value(Request(None, 'POST'))
            return

        # <form>  (initialize the new form object)
        if self._form_tag.begin(tag):
            form_id = attrs.get('id') or attrs.get('name')
            self._form_tag.set_value(Request(attrs.get('action'), attrs.get('method'), form_id))
            return

        # <form> ... </form>  (fill the new form object with parameters)
        if self._form_tag.is_processing():
            # extract form parameters
            if tag == 'input' and attrs.get('type') in ['hidden', 'text', 'password'] and attrs.get('name'):
                self._form_tag.value.add_param(attrs['name'], attrs.get('value', ''))
            # if form has no action attribute, try to use one from the submit button
            if not self._form_tag.value.is_complete():
                self._form_tag.value.process_action(tag, attrs)
        # <body> ... </body>  (uhh, get buttons outside <form>)
        elif self._body_tag.is_processing():
            if not self._body_tag.value.is_complete():
                self._body_tag.value.process_action(tag, attrs)

    def handle_endtag(self, tag):
        # </form>  (copy completed form into list)
        if self._form_tag.end(tag):
            if self._form_tag.value.is_complete():
                self.forms.append(self._form_tag.value)
                self._form_tag.clear_value()
        # </body>
        elif self._body_tag.end(tag):
            if self._body_tag.value.is_complete():
                self.global_action = self._body_tag.value
        # </noscript>
        elif self._noscript_tag.end(tag):
            pass
        # </script>
        elif self._script_tag.end(tag):
            pass
        # </html>  (finish HTML processing, perform all onexit steps)
        elif tag == 'html':
            if self._script_tag.value:
                match = reScriptLocation.search(self._script_tag.value)
                if match:
                    self.script = Request(match.group(1))

    def handle_data(self, data):
        # <script> body
        if self._script_tag.is_processing():
            self._script_tag.set_value(data)

###############################################################################
# Class for retrieving HTML page actions
#
class PageActionsExtractor:
    def __init__(self, body):
        self.pp = None
        self.body = body

        try:
            self.pp = PageParser()
            self.pp.feed(body)
        except Exception as ex:
            logging.debug("Page not completely processed due to: %s", ex)
            logging.debug("see corresponded html in file: %s", log2file(body))
            # do not raise here, because it happens quite often on some complex pages

    def get_action(self, uid=None):
        # Search for action with specific uid and returns Request object corresponded this action.
        # If uid is None, the auto-redirected action is searching, accordingly hardcoded rules.
        # If no any actions found, None is returned.
        #
        # uid -- can be either <form id> or <form button id>; in later case, the uid has
        # 'button:' prefix (e.g. for button id='login', uid='button:login')

        if not self.pp:
            return None

        if not uid:  # looking for auto-redirected request
            if self.pp.script:
                return self.pp.script

            if self.pp.iframe_src:
                return self.pp.iframe_src

            if self.pp.forms and self.pp.form_auto_submit:
                return self.pp.forms[0]

            if self.pp.body_onload:
                return self.pp.body_onload
        else:  # Looking for particular request
            if self.pp.forms:
                for form in self.pp.forms:
                    if form.uid == uid:
                        return form
            if self.pp.global_action:
                if self.pp.global_action.uid == uid:
                    return self.pp.global_action

        return None

###############################################################################
# Helper functions
#
def unescape(s):
    s = s.replace("&lt;", "<")
    s = s.replace("&gt;", ">")
    s = s.replace("&quot;", '"')
    s = s.replace("&apos;", "'")
    s = s.replace("&amp;", "&")
    return s


def log2file(text):
    """
    Pretty simple function, saves the text into temporary file and returns it's name
    :param text: string to log
    :return: temporary file name
    """
    handle, name = tempfile.mkstemp(suffix='.html', prefix='log_', text=True)
    os.write(handle, text)
    os.close(handle)
    return name

class HTTPRequestFromStr(BaseHTTPRequestHandler):
    def __init__(self, request_text):
        self.rfile = StringIO(request_text)
        self.raw_requestline = self.rfile.readline()
        self.error_code = self.error_message = None
        self.parse_request()

    def send_error(self, code, message):
        self.error_code = code
        self.error_message = message

class HTTPResponseFromStr(HTTPResponse):
    def __init__(self, response_text):
        try:
            self.init(response_text)
        except httplib.UnknownProtocol as e:
            # FIXME: handle HTTP/2.0 that is 'unknown' for httplib (to check go to www.google.com)
            logging.debug(str(e) + " retrying with HTTP/2.0 replaced by HTTP/1.1")
            self.init(response_text.replace("HTTP/2.0", "HTTP/1.1"))

    def init(self, response_text):
        class FakeSocket:
            def __init__(self, response_str):
                self._file = StringIO(response_str)
            def makefile(self, *args, **kwargs):
                return self._file

        source = FakeSocket(response_text)
        HTTPResponse.__init__(self, source)
        self.begin()



#!/usr/bin/env python

from __future__ import print_function, absolute_import

# -*- coding: utf-8 -*-
__author__ = "perfguru87@gmail.com"
__copyright__ = "Copyright 2018, The PerfTracker project"
__license__ = "MIT"

"""
The Firefox browser helper
"""

import logging
import re
import os
import datetime
import time
import json
import sys

from perftrackerlib.helpers.timeparser import TimeParser
from perftrackerlib.helpers.largelogfile import LargeLogFile
from perftrackerlib.helpers import timehelpers

from .utils import parse_url, get_val
from .browser_webdriver import BrowserWebdriver, abort
from .browser_base import BrowserExc
from .page import PageRequest, PageTimeline
from . import httputils

try:
    from selenium import webdriver
    from selenium.common.exceptions import WebDriverException
    import psutil
except ImportError as e:
    abort(e)


# reFF = re.compile(r"(\d+-\d+-\d+ \d+:\d+\:\d+.\d+) UTC - [-\d]+\[([\d\w]+)\]:\s+(.*)\n")
reFF = re.compile(r" UTC - [-\d]+\[([\d\w]+)\]:\s+(.*)\n")
reFFTxnInit = re.compile(r"^nsHttpTransaction::Init \[this=([\d\w]+).*$")
reFFProcessData = re.compile(r"^nsHttpTransaction::ProcessData \[this=([\d\w]+) count=(\d+)\].*$")
reFFReadFromcache = re.compile(r"^nsHttpChannel::ReadFromCache \[this=([\d\w]+)] Using cached copy of: (.*)$")
reFFDestroyTxn = re.compile(r"^Destroying nsHttpTransaction @([\d\w]+).*$")


class BrowserFirefox(BrowserWebdriver):
    engine = "firefox"
    skip_urls = ["cdn.mozilla.net", "mozilla.org", "mozilla.com", "ocsp.digicert.com", "openh264.org"]

    def __init__(self, *args, **kwargs):
        BrowserWebdriver.__init__(self, *args, **kwargs)
        self._last_seen_dt = datetime.datetime.utcnow()

    def _skip_url(self, page, url):
        if url and url.startswith("about:"):
            return True

        return BrowserWebdriver._skip_url(self, page, url)

    def _browser_get_rss_kb(self):
        return psutil.Process(self.pid).memory_info().rss / 1024

    def browser_start(self):
        os.environ['NSPR_LOG_FILE'] = self.log_path
        os.environ['NSPR_LOG_MODULES'] = 'timestamp,nsHttp:5,nsSocketTransport:5,nsStreamPump:5,nsHostResolver:5'

        profile = webdriver.FirefoxProfile()
        profile.accept_untrusted_certs = True

        # sometimes (bug? race?) firefox fails to start wihout clear reason, so lets try to restart it several times
        retry = 5
        while True:
            try:
                self.driver = webdriver.Firefox(firefox_profile=profile)
            except WebDriverException as e:
                if retry:
                    self.log_warning("firefox start failed, retrying...")
                    time.sleep(1)
                    retry -= 1
                    continue
                self.log_error("Firefox browser start failed, see browser logs here: %s" % self.log_path)
                if "Can't load the profile" in str(e):
                    # http://stackoverflow.com/questions/6682009/selenium-firefoxprofile-exception-cant-load-the-profile
                    self.log_error("Try to update Firefox and selenium: pip install -U selenium")
                if "geckodriver" in str(e):
                    self.log_error("Try to install geckodriver from https://github.com/mozilla/geckodriver/releases and"
                                   " put it to PATH")
                raise BrowserExc(str(e))
            break

        # old native firefox driver
        if self.driver.binary:
            return self.driver.binary.process.pid

        # geckodriver
        return self.driver.service.process.pid

    def navigation_reset(self):
        self.navigate_to("about:buildconfig")

    def _browser_get_events(self, page):
        thread2req = {}
        thread2resp = {}
        thread2txn = {}
        path2prot = {}

        tp = TimeParser()

        start = self._last_seen_dt

        self.log_debug("Reading %s starting from %s" % (self.log_path, start))

        log = LargeLogFile(self.log_path,
                           start.strftime("%Y-%m-%d %H:%M:%S.000000"),
                           "2099-01-01 00:00:00.000000")

        for dt, line in log.readlines_with_time():

            if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
                self.log_debug("LOG: %s" % line.strip())

            m = reFF.match(line)
            if not m:
                continue

            self._last_seen_dt = dt

            ts = timehelpers.dt2ts_utc(dt) * 1000
            thread = m.group(2)
            msg = m.group(3)

            m = reFFTxnInit.match(msg)
            if m:
                thread2txn[thread] = m.group(1)
                continue

            m = reFFProcessData.match(msg)
            if m:
                txn = m.group(1)
                thread2txn[thread] = txn
                req = page.get_request(txn)
                if req:
                    req.length += int(m.group(2))
                continue

            if msg.startswith("uri="):
                url = msg[4:]
                prot, domain, path = parse_url(url)
                path2prot["%s%s" % (domain, path)] = prot

            if msg.startswith("nsHttpChannel::ReadFromCache"):
                m = reFFReadFromcache.match(msg)
                if m:
                    txn = m.group(1)
                    r = PageRequest(page, txn)
                    r.method = "GET"
                    r.url = m.group(2)
                    r.status = 200
                    r.start(ts)
                    r.complete(ts)
                    r.cached = True
                    page.process_activity("read request from cache", ts)
                    page.add_request(r)

            m = reFFDestroyTxn.match(msg)
            if m:
                txn = m.group(1)
                r = page.get_request(txn)
                if r and not r.completed:
                    self.log_debug("WARNING: request '%s' has been completed abnormally" % r.id)
                    r.complete(ts)
                continue

            if thread not in thread2txn:
                continue

            txn = thread2txn[thread]

            if msg.startswith("http request"):
                thread2req[thread] = []
                page.add_request(PageRequest(page, txn))
                continue

            if thread in thread2req:
                if msg == "]":
                    req = page.get_request(txn)
                    r = httputils.HTTPRequestFromStr("\n".join(thread2req[thread]))
                    req.method = r.command

                    p = "%s%s" % (r.headers['host'], r.path)
                    _, domain, path = parse_url("http://" + p)
                    path = "%s%s" % (domain, path)

                    skip = False

                    if self._skip_url(page, path):
                        skip = True
                        self.log_debug("ignore/skip request to: %s" % path)
                    elif path in path2prot:
                        req.url = "%s://%s" % (path2prot[path], p)
                    else:
                        self.log_warning("Can't determine protocol for url: %s, skipping!" % p)
                        skip = True

                    if skip:
                        del thread2req[thread]
                        page.del_request(req)
                        continue

                    req.header = r.headers.dict

                    del thread2req[thread]
                    page.process_activity("request", ts)
                    req.start(ts)
                else:
                    thread2req[thread].append(msg)

            if msg.startswith("http response"):
                if page.get_request(txn):
                    thread2resp[thread] = []
                    continue

            if thread in thread2resp:
                if msg == "]":
                    req = page.get_request(txn)
                    if not req:
                        continue
                    r = httputils.HTTPResponseFromStr("\n".join(thread2resp[thread]))
                    # print("\n".join(thread2resp[thread]))
                    len = r.getheader('Content-Length')
                    if len:
                        req.content_length = int(len)
                    req.status = r.status
                    type = r.getheader("Content-Type")
                    if type:
                        req.set_type(type)
                    req.keepalive = r.getheader("Connection") in ("Keep-Alive", "keep-alive")
                    req.gzipped = r.getheader("Content-Encoding") == "gzip"

                    del thread2resp[thread]
                    page.process_activity("response", ts)
                    req.complete(ts)
                else:
                    thread2resp[thread].append(msg)


##############################################################################
# Autotests
##############################################################################


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    try:
        b = BrowserFirefox()
    except BrowserExc as e:
        sys.exit(-1)
    b.navigate_to("https://example.com/")
    b.browser_stop()



"""
perftracker-cp-crawler
https://packaging.python.org/en/latest/distributing.html
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

import perftracker_cp_crawler

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='perftracker-cp-crawler',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=perftracker_cp_crawler.__version__,

    description='Python package with the perftracker control panels crawler',
    long_description=long_description,
    long_description_content_type='text/markdown',

    # Author details
    author='perfguru87',
    author_email='perfguru87@gmail.com',
    url='http://www.perftracker.org',

    # Choose your license
    license='MIT License',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    # What does your project relate to?
    keywords='performance tests tools selenium webdriver perftracker client',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=open(path.join(path.dirname(__file__), 'requirements.txt'), 'r').readlines(),

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={'test': ['pycodestyle', 'coverage'], },
    scripts=['examples/pt-wp-crawler.py']
)
